/*      ---------- Mini-Scheme Interpreter Version 0.85 ----------
 *                coded by Atsushi Moriwaki (11/5/1989)
 *  This version has been modified by Chris Pressey.
 *    current version is 0.85p1 (as yet unreleased)
 *  This version has been modified by R.C. Secrist.
 *  This is a revised and modified version by Akira KIDA.
 
 *               THIS SOFTWARE IS IN THE PUBLIC DOMAIN
 *               ------------------------------------
*/
// only for Windows, tested with Win10
// added TOK_EOF, EOF_OBJ, OP_GENSYM, some modifications 
// works with "msinit.scm" 

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

// #define VERBOSE    /* define this if you want verbose GC */    
# define BACKQUOTE '`'    // 96, 0x60

/*
 *  Basic memory allocation units
 */
#define CELL_SEGSIZE    5000   /* # of cells in one segment */
#define CELL_NSEGMENT   100    /* # of segments for cells */
#define STR_SEGSIZE     2500   /* bytes of one string segment */
#define STR_NSEGMENT    100    /* # of segments for strings */

#define banner "Welcome to Mini-Scheme Version 0.85 mod.\n"

#define prompt "> "
#define InitFile "msinit.scm"
#define FIRST_CELLSEGS 3

/* cell structure */
struct cell {
    unsigned short _flag;
    union {
        struct {
            char   *_svalue;
            short   _keynum;
        } _string;
        struct {
            long    _ivalue;
        } _number;
        struct {
            struct cell *_car;
            struct cell *_cdr;
        } _cons;
    } _object;
};

typedef struct cell *pointer;

#define T_STRING         1    /* 0000000000000001 */
#define T_NUMBER         2    /* 0000000000000010 */
#define T_SYMBOL         4    /* 0000000000000100 */
#define T_SYNTAX         8    /* 0000000000001000 */
#define T_PROC          16    /* 0000000000010000 */
#define T_PAIR          32    /* 0000000000100000 */
#define T_CLOSURE       64    /* 0000000001000000 */
#define T_CONTINUATION 128    /* 0000000010000000 */
# define T_MACRO       256    /* 0000000100000000 */
#define T_PROMISE      512    /* 0000001000000000 */
#define T_ATOM       16384    /* 0100000000000000 */    /* only for gc */

#define CLRATOM_     49151    /* 1011111111111111 */    /* only for gc */
#define MARK_        32768    /* 1000000000000000 */
#define UNMARK       32767    /* 0111111111111111 */

/* macros for cell operations */
#define type_(p)         ((p)->_flag)

#define isstring(p)     (type_(p)&T_STRING)
#define strvalue(p)     ((p)->_object._string._svalue)
#define keynum(p)       ((p)->_object._string._keynum)

#define isnumber(p)     (type_(p)&T_NUMBER)
#define ivalue(p)       ((p)->_object._number._ivalue)

#define ispair(p)       (type_(p)&T_PAIR)
#define car(p)          ((p)->_object._cons._car)
#define cdr(p)          ((p)->_object._cons._cdr)

#define issymbol(p)     (type_(p)&T_SYMBOL)
#define symname(p)      strvalue(car(p))
#define hasprop(p)      (type_(p)&T_SYMBOL)
#define symprop(p)      cdr(p)

#define issyntax(p)     (type_(p)&T_SYNTAX)
#define isproc(p)       (type_(p)&T_PROC)
#define syntaxname(p)   strvalue(car(p))
#define syntaxnum(p)    keynum(car(p))
#define procnum(p)      ivalue(p)

#define isclosure(p)    (type_(p)&T_CLOSURE)
#define ismacro(p)      (type_(p)&T_MACRO)
#define closure_code(p) car(p)
#define closure_env(p)  cdr(p)

#define iscontinuation(p) (type_(p)&T_CONTINUATION)
#define cont_dump(p)      cdr(p)

#define ispromise(p)    (type_(p)&T_PROMISE)
#define setpromise(p)   type_(p) |= T_PROMISE

#define isatom(p)       (type_(p)&T_ATOM)
#define setatom(p)      type_(p) |= T_ATOM
#define clratom(p)      type_(p) &= CLRATOM_

#define ismark(p)       (type_(p)&MARK_)
#define setmark(p)      type_(p) |= MARK_
#define clrmark(p)      type_(p) &= UNMARK

#define caar(p)         car(car(p))
#define cadr(p)         car(cdr(p))
#define cdar(p)         cdr(car(p))
#define cddr(p)         cdr(cdr(p))
#define cadar(p)        car(cdr(car(p)))
#define caddr(p)        car(cdr(cdr(p)))
#define cadaar(p)       car(cdr(car(car(p))))
#define cadddr(p)       car(cdr(cdr(cdr(p))))
#define cddddr(p)       cdr(cdr(cdr(cdr(p))))

/* arrays for segments */
pointer cell_seg[CELL_NSEGMENT];
int last_cell_seg = -1;
char *str_seg[STR_NSEGMENT];
int str_seglast = -1;

// automatic symbol generator
unsigned long   sym_count;

/* We use 4 registers. */
pointer args;            /* register for arguments of function */
pointer envir;           /* stack register for current environment */
pointer code;            /* register for current code */
pointer dump;            /* stack register for next evaluation */

struct cell _NIL;
pointer NIL = &_NIL;    /* special cell representing empty cell */
struct cell _T;
pointer T = &_T;        /* special cell representing #t */
struct cell _F;
pointer F = &_F;        /* special cell representing #f */

struct cell _EOF_OBJ;
pointer EOF_OBJ = &_EOF_OBJ;	/* special cell representing end-of-file */

pointer oblist = &_NIL;         /* pointer to symbol table */
pointer global_env;             /* pointer to global environment */

/* global pointers to special symbols */
pointer LAMBDA;            /* pointer to syntax lambda */
pointer QUOTE_;            /* pointer to syntax quote */

pointer QQUOTE;            /* pointer to symbol quasiquote */
pointer UNQUOTE;           /* pointer to symbol unquote */
pointer UNQUOTESP;         /* pointer to symbol unquote-splicing */

pointer free_cell = &_NIL; /* pointer to top of free cells */
long    fcells = 0;        /* # of free cells */

FILE   *infp;            /* input file */
FILE   *outfp;            /* output file */

char    gc_verbose;             /* if gc_verbose is not zero, print gc status */
int     all_errors_fatal = 0;   /* if not zero, every error is a FatalError */

void gc(pointer *a,  pointer *b);
void FatalError(char *s);
void init_globals();
int isdelim(char *s, char c);

/* allocate new cell segment */
int alloc_cellseg(int n)
{
     pointer p;
     long i;
     int k;

    for (k = 0; k < n; k++) {
        if (last_cell_seg >= CELL_NSEGMENT - 1)
            return k;
        p = (pointer) malloc(CELL_SEGSIZE * sizeof(struct cell));
        if (p == (pointer) 0)
            return k;
        cell_seg[++last_cell_seg] = p;
        fcells += CELL_SEGSIZE;
        for (i = 0; i < CELL_SEGSIZE - 1; i++, p++) {
            type_(p) = 0;
            car(p) = NIL;
            cdr(p) = p + 1;
        }
        type_(p) = 0;
        car(p) = NIL;
        cdr(p) = free_cell;
        free_cell = cell_seg[last_cell_seg];
    }
    return n;
}

/* allocate new string segment */
int alloc_strseg(int n)
{
     char *p;
     long i;
     int k;

    for (k = 0; k < n; k++) {
        if (str_seglast >= STR_NSEGMENT)
            return k;
        p = (char *) malloc(STR_SEGSIZE * sizeof(char));
        if (p == (char *) 0)
            return k;
        str_seg[++str_seglast] = p;
        for (i = 0; i < STR_SEGSIZE; i++)
            *p++ = (char) (-1);
    }
    return n;
}

/* initialization of Mini-Scheme */
void init_scheme()
{
    if (alloc_cellseg(FIRST_CELLSEGS) != FIRST_CELLSEGS)
        FatalError("Unable to allocate initial cell segments");
    if (!alloc_strseg(1))
        FatalError("Unable to allocate initial string segments");
#ifdef VERBOSE
    gc_verbose = 1;
#else
    gc_verbose = 0;
#endif
    init_globals();
}

/* get new cell.  parameter a, b is marked by gc. */
pointer get_cell(pointer *a,  pointer *b)
{
     pointer x;

    if (free_cell == NIL) {
        gc(a, b);
        if (free_cell == NIL)
            if (!alloc_cellseg(1))
                FatalError("run out of cells  --- unable to recover cells");
    }
    x = free_cell;
    free_cell = cdr(x);
    --fcells;
    return (x);
}

/* get new cons cell */
pointer cons(pointer a,  pointer b)
{
     pointer x = get_cell(&a, &b);

    type_(x) = T_PAIR;
    car(x) = a;
    cdr(x) = b;
    return (x);
}

/* get number atom */
pointer mk_number(long num)
{
     pointer x = get_cell(&NIL, &NIL);

    type_(x) = (T_NUMBER | T_ATOM);
    ivalue(x) = num;
    return (x);
}

/* allocate name to string area */
char *store_string(char *name)
{
     char *q;
     short i;
    long    len, remain;

    /* first check name has already listed */
    for (i = 0; i <= str_seglast; i++)
        for (q = str_seg[i]; *q != (char) (-1); ) {
            if (!strcmp(q, name))
                goto FOUND;
            while (*q++)
                ;    /* get next string */
        }
    len = strlen(name) + 2;
    remain = (long) STR_SEGSIZE - ((long) q - (long) str_seg[str_seglast]);
    if (remain < len) {
        if (!alloc_strseg(1))
            FatalError("run out of string area");
        q = str_seg[str_seglast];
    }
    strcpy(q, name);
FOUND:
    return (q);
}

/* get new string */
pointer mk_string(char *str)
{
     pointer x = get_cell(&NIL, &NIL);

    strvalue(x) = store_string(str);
    type_(x) = (T_STRING | T_ATOM);
    keynum(x) = (short) (-1);
    return (x);
}

/* get new symbol */
pointer mk_symbol(char *name)
{
     pointer x;

    /* first check oblist */
    for (x = oblist; x != NIL; x = cdr(x))
        if (!strcmp(name, symname(car(x))))
            break;

    if (x != NIL)
        return (car(x));
    else {
        x = cons(mk_string(name), NIL);
        type_(x) = T_SYMBOL;
        oblist = cons(x, oblist);
        return (x);
    }
}

// automatic symbol generator
pointer auto_gen(void)
{
    char newsym[40];
    sym_count++;
    sprintf(newsym, "autosym_%lu", sym_count);
    return mk_symbol(newsym);
}

/* make symbol or number atom from string */
pointer mk_atom(char *q)
{
    char    c, *p;

    p = q;
    if (!isdigit(c = *p++)) {
        if ((c != '+' && c != '-') || !isdigit(*p))
            return (mk_symbol(q));
    }
    for ( ; (c = *p) != 0; ++p)
        if (!isdigit(c))
            return (mk_symbol(q));
    return (mk_number(atol(q)));
}

/* make constant */
pointer mk_const(char *name)
{
    long    x;
    char    tmp[256];

    if (!strcmp(name, "t"))
        return (T);
    else if (!strcmp(name, "f"))
        return (F);
    else if (*name == 'o') {/* #o (octal) */
        sprintf(tmp, "0%s", &name[1]);
        sscanf(tmp, "%lo", (unsigned long int *)&x);
        return (mk_number(x));
    } else if (*name == 'd') {    /* #d (decimal) */
        sscanf(&name[1], "%ld", &x);
        return (mk_number(x));
    } else if (*name == 'x') {    /* #x (hex) */
        sprintf(tmp, "0x%s", &name[1]);
        sscanf(tmp, "%lx", (unsigned long int *)&x);
        return (mk_number(x));
    } else
        return (NIL);
}


/* ========== garbage collector ========== */

/*--
 *  We use algorithm E (Kunuth, The Art of Computer Programming Vol.1,
 *  sec.3.5) for marking.
 */
void mark(pointer a)
{
     pointer t, q, p;

E1: t = (pointer) 0;
    p = a;
E2: setmark(p);
E3: if (isatom(p))
        goto E6;
E4: q = car(p);
    if (q && !ismark(q)) {
        setatom(p);
        car(p) = t;
        t = p;
        p = q;
        goto E2;
    }
E5: q = cdr(p);
    if (q && !ismark(q)) {
        cdr(p) = t;
        t = p;
        p = q;
        goto E2;
    }
E6: if (!t)
        return;
    q = t;
    if (isatom(q)) {
        clratom(q);
        t = car(q);
        car(q) = p;
        p = q;
        goto E5;
    } else {
        t = cdr(q);
        cdr(q) = p;
        p = q;
        goto E6;
    }
}


/* garbage collection. parameter a, b is marked. */
void gc(pointer *a,  pointer *b)
{
     pointer p;
     short i;
     long j;

    if (gc_verbose)
        printf("gc...");

    /* mark system globals */
    mark(oblist);
    mark(global_env);

    /* mark current registers */
    mark(args);
    mark(envir);
    mark(code);
    mark(dump);

    /* mark variables a, b */
    mark(*a);
    mark(*b);

    /* garbage collect */
    clrmark(NIL);
    fcells = 0;
    free_cell = NIL;
    for (i = 0; i <= last_cell_seg; i++) {
        for (j = 0, p = cell_seg[i]; j < CELL_SEGSIZE; j++, p++) {
            if (ismark(p))
                clrmark(p);
            else {
                type_(p) = 0;
                cdr(p) = free_cell;
                car(p) = NIL;
                free_cell = p;
                ++fcells;
            }
        }
    }

    if (gc_verbose)
        printf(" done %ld cells are recovered.\n", fcells);
}


/* ========== Routines for Reading ========== */
#define TOK_EOF     (-1)

#define TOK_LPAREN  0
#define TOK_RPAREN  1
#define TOK_DOT     2
#define TOK_ATOM    3
#define TOK_QUOTE   4
#define TOK_COMMENT 5
#define TOK_DQUOTE  6
# define TOK_BQUOTE  7
# define TOK_COMMA   8
# define TOK_ATMARK  9
#define TOK_SHARP   10

//#define EOF      0xFF // -1
#define LINESIZE 1024

char    linebuff[LINESIZE];
char    strbuff[256];
char   *currentline = linebuff;
char   *endline = linebuff;

/* get new character from input file */
int inchar()
{
    if (currentline >= endline) {    /* input buffer is empty */
        if (feof(infp)) {
            fclose(infp);
            infp = stdin;
            //printf(prompt);
            return EOF;
        }
        strcpy(linebuff, "\n");
        if (fgets(currentline = linebuff, LINESIZE, infp) == NULL)
            if (infp == stdin) {     //^C
                fprintf(stderr, "Good-bye\n");
                exit(0);
            }
        endline = linebuff + strlen(linebuff);
    }
    return (*currentline++);
}

/* clear input buffer */
void clearinput()
{
    currentline = endline = linebuff;
}

/* back to standard input */
flushinput()
{
    if (infp != stdin) {
        fclose(infp);
        infp = stdin;
    }
    clearinput();
}

/* back character to input buffer */
void backchar()
{
    currentline--;
}

/* read chacters to delimiter */
char *readstr(char *delim)
{
    char   *p = strbuff;

    while (isdelim(delim, (*p++ = inchar())))
        ;
    backchar();
    *--p = '\0';
    return (strbuff);
}

/* read string expression "xxx...xxx" */
char *readstrexp()
{
    char c, *p = strbuff;

    for (;;) {
        if ((c = inchar()) != '"')
            *p++ = c;
        else if (p > strbuff && *(p - 1) == '\\')
            *(p - 1) = '"';
        else {
            *p = '\0';
            return (strbuff);
        }
    }
}

/* check c is delimiter */
int isdelim(char *s, char c)
{
	if (c == EOF) return 0;
	while (*s)
		if (*s++ == c)
			return 0;
	return 1;
}

/* skip white characters */
int skipspace()
{
    int c = inchar();
    while (isspace(c))
        c = inchar();
    if (c != EOF)
    {    
      backchar();
      return 1;
    }
    return EOF;  
}

/* get token */
int token()
{
//    skipspace();
	int c = skipspace();
	if (c == EOF) {
		return TOK_EOF;
	}

    switch (inchar()) {
      case EOF:
        return TOK_EOF;
    
    case '(':
        return (TOK_LPAREN);
    case ')':
        return (TOK_RPAREN);
    case '.':
        return (TOK_DOT);
    case '\'':
        return (TOK_QUOTE);
    case ';':     // ';' ==> bypass comments recursively altogether
//        return (TOK_COMMENT);
        c = inchar();
        while ((c != '\n') && (c != EOF)) // '\n'
          // skip comments
          c = inchar();

        if (c == EOF)
            return TOK_EOF;
        return token();

    case '"':
        return (TOK_DQUOTE);
    case BACKQUOTE:
        return (TOK_BQUOTE);
    case ',':
        if (inchar() == '@')
            return (TOK_ATMARK);
        else {
            backchar();
            return (TOK_COMMA);
        }
    case '#':
        return (TOK_SHARP);
    default:
        backchar();
        return (TOK_ATOM);
    }
}

/* ========== Routines for Printing ========== */
#define  ok_abbrev(x)    (ispair(x) && cdr(x) == NIL)

void strunquote(char *p, char *s)
{
    *p++ = '"';
    for ( ; *s; ++s) {
        if (*s == '"') {
            *p++ = '\\';
            *p++ = '"';
        } else if (*s == '\n') {
            *p++ = '\\';
            *p++ = 'n';
        } else
            *p++ = *s;
    }
    *p++ = '"';
    *p = '\0';
}


/* print atoms */
int printatom(pointer l, int f)
{
    char    *p;
    
    if (l == NIL)
        p = "()";
    else if (l == T)
        p = "#t";
    else if (l == F)
        p = "#f";
    else if (l == EOF_OBJ)
        p = "#<EOF>";    
    else if (isnumber(l)) {
        p = strbuff;
        sprintf(p, "%ld", ivalue(l));
    } else if (isstring(l)) {
        if (!f)
            p = strvalue(l);
        else {
            p = strbuff;
            strunquote(p, strvalue(l));
        }
    } else if (issymbol(l))
        p = symname(l);
    else if (isproc(l)) {
        p = strbuff;
        sprintf(p, "#<PROCEDURE %ld>", procnum(l));
    } else if (ismacro(l)) {
        p = "#<MACRO>";
    } else if (isclosure(l))
        p = "#<CLOSURE>";
    else if (iscontinuation(l))
        p = "#<CONTINUATION>";
    if (f < 0)
        return strlen(p);
    fputs(p, outfp);
    return 0;
}


/* ========== Routines for Evaluation Cycle ========== */

/* make closure. c is code. e is environment */
pointer mk_closure(pointer c,  pointer e)
{
     pointer x = get_cell(&c, &e);

    type_(x) = T_CLOSURE;
    car(x) = c;
    cdr(x) = e;
    return (x);
}

/* make continuation. */
pointer mk_continuation(pointer d)
{
     pointer x = get_cell(&NIL, &d);

    type_(x) = T_CONTINUATION;
    cont_dump(x) = d;
    return (x);
}

/* reverse list -- make new cells */
pointer reverse(pointer a)     /* a must be checked by gc */
{
    pointer p = NIL;

    for ( ; ispair(a); a = cdr(a))
        p = cons(car(a), p);
    return (p);
}

/* reverse list --- no make new cells */
pointer non_alloc_rev(pointer term, pointer list)
{
    pointer p = list, result = term, q;

    while (p != NIL) {
        q = cdr(p);
        cdr(p) = result;
        result = p;
        p = q;
    }
    return (result);
}

/* append list -- make new cells */
pointer append(pointer a,  pointer b)
{
    pointer p = b, q;

    if (a != NIL) {
        a = reverse(a);
        while (a != NIL) {
            q = cdr(a);
            cdr(a) = p;
            p = a;
            a = q;
        }
    }
    return (p);
}

/* equivalence of atoms */
int eqv(pointer a,  pointer b)
{
    if (isstring(a)) {
        if (isstring(b))
            return (strvalue(a) == strvalue(b));
        else
            return (0);
    } else if (isnumber(a)) {
        if (isnumber(b))
            return (ivalue(a) == ivalue(b));
        else
            return (0);
    } else
        return (a == b);
}

/* true or false value macro */
// #define istrue(p)       ((p) != NIL && (p) != F)
// #define isfalse(p)      ((p) == NIL || (p) == F)
#define istrue(p)       ((p) != F)
#define isfalse(p)      ((p) == F)

/* Error macro */
#define Error_0(s) do {                       \
    args = cons(mk_string((s)), NIL);          \
    operator = (short)OP_ERR0;                 \
    return T; } while (0)

#define Error_1(s, a) do {                    \
    args = cons((a), NIL);                     \
    args = cons(mk_string((s)), args);         \
    operator = (short)OP_ERR0;                 \
    return T; } while (0)

/* control macros for Eval_Cycle */
#define s_goto(a) do {                        \
    operator = (short)(a);                     \
    return T; } while (0)

#define s_save(a, b, c)  (                     \
    dump = cons(envir, cons((c), dump)),       \
    dump = cons((b), dump),                    \
    dump = cons(mk_number((long)(a)), dump))   \


#define s_return(a) do {                      \
    value = (a);                               \
    operator = ivalue(car(dump));              \
    args = cadr(dump);                         \
    envir = caddr(dump);                       \
    code = cadddr(dump);                       \
    dump = cddddr(dump);                       \
    return T; } while (0)

#define s_retbool(tf)    s_return((tf) ? T : F)


/* ========== Evaluation Cycle ========== */

/* operator code */
enum {
    OP_LOAD      =  0,
    OP_T0LVL,
    OP_T1LVL,
    OP_READ,
    OP_VALUEPRINT,
    OP_EVAL,
    OP_E0ARGS,
    OP_E1ARGS,
    OP_APPLY,
    OP_DOMACRO,
    OP_GENSYM,

    OP_LAMBDA,       // 11
    OP_QUOTE,
    OP_DEF0,
    OP_DEF1,
    OP_BEGIN,
    OP_IF0,
    OP_IF1,
    OP_SET0,
    OP_SET1,
    OP_LET0,
    OP_LET1,
    OP_LET2,
    OP_LET0AST,
    OP_LET1AST,
    OP_LET2AST,

    OP_LET0REC,      // 26
    OP_LET1REC,
    OP_LET2REC,
    OP_COND0,
    OP_COND1,
    OP_DELAY,
    OP_AND0,
    OP_AND1,
    OP_OR0,
    OP_OR1,
    OP_C0STREAM,
    OP_C1STREAM,
    OP_0MACRO,
    OP_1MACRO,
    OP_CASE0,
    OP_CASE1,
    OP_CASE2,

    OP_PEVAL,        // 43
    OP_PAPPLY,
    OP_CONTINUATION,

    OP_ADD,          // 46
    OP_SUB,
    OP_MUL,
    OP_DIV,
    OP_REM,
    OP_CAR,
    OP_CDR,
    OP_CONS,
    OP_SETCAR,
    OP_SETCDR,

    OP_NOT,          // 56
    OP_BOOL,
    OP_NULL,
    OP_ZEROP,
    OP_POSP,
    OP_NEGP,
    OP_NEQ,
    OP_LESS,
    OP_GRE,
    OP_LEQ,
    OP_GEQ,
    OP_SYMBOL,
    OP_NUMBER,
    OP_STRING,
    OP_PROC,
    OP_PAIR,
    OP_EQ,
    OP_EQV,

    OP_FORCE,        // 74
    OP_WRITE,
    OP_DISPLAY,
    OP_NEWLINE,
    OP_ERR0,
    OP_ERR1,
    OP_REVERSE,
    OP_APPEND,
    OP_PUT,
    OP_GET,
    OP_QUIT,
    OP_GC,
    OP_GCVERB,
    OP_NEWSEGMENT,

    OP_RDSEXPR,      // 88
    OP_RDLIST,
    OP_RDDOT,
    OP_RDQUOTE,
    OP_RDQQUOTE,
    OP_RDUNQUOTE,
    OP_RDUQTSP,
    OP_P0LIST,
    OP_P1LIST,

    OP_LIST_LENGTH,  // 97
    OP_ASSQ,
    OP_PRINT_WIDTH,
    OP_P0_WIDTH,
    OP_P1_WIDTH,
    OP_GET_CLOSURE,
    OP_CLOSUREP,
    OP_MACROP,       // 104
};

static FILE *tmpfp;
static int tok;
static int print_flag;
static pointer value;
static short operator;

pointer opexe_0(short op)
{
    pointer x, y;

    switch (op) {
    case OP_LOAD:        /* load */
        if (!isstring(car(args))) {
            Error_0("load -- argument is not string");
        }
        if ((infp = fopen(strvalue(car(args)), "r")) == NULL) {
            infp = stdin;
            Error_1("Unable to open", car(args));
        }
        fprintf(outfp, "loading %s", strvalue(car(args)));
        s_goto(OP_T0LVL);

    case OP_T0LVL:    /* top level */
        fprintf(outfp, "\n");
        dump = NIL;
        envir = global_env;
        s_save(OP_VALUEPRINT, NIL, NIL);
        s_save(OP_T1LVL, NIL, NIL);
        if (infp == stdin)
            printf(prompt);
        s_goto(OP_READ);

    case OP_T1LVL:    /* top level */
        code = value;
        s_goto(OP_EVAL);
        
    case OP_READ:        /* read */
        tok = token();
        s_goto(OP_RDSEXPR);

    case OP_VALUEPRINT:    /* print evalution result */
        print_flag = 1;
        args = value;
        s_save(OP_T0LVL, NIL, NIL);
        s_goto(OP_P0LIST);

    case OP_EVAL:        /* main part of evalution */
        if (issymbol(code)) {    /* symbol */
            for (x = envir; x != NIL; x = cdr(x)) {
                for (y = car(x); y != NIL; y = cdr(y))
                    if (caar(y) == code)
                        break;
                if (y != NIL)
                    break;
            }
            if (x != NIL) {
                s_return(cdar(y));
            } else {
                Error_1("Unbounded variable", code);
            }
        } else if (ispair(code)) {
            if (issyntax(x = car(code))) {    /* SYNTAX */
                code = cdr(code);
                s_goto(syntaxnum(x));
            } else {/* first, eval top element and eval arguments */
                s_save(OP_E0ARGS, NIL, code);
                code = car(code);
                s_goto(OP_EVAL);
            }
        } else {
            s_return(code);
        }

    case OP_E0ARGS:    /* eval arguments */
        if (ismacro(value)) {    /* macro expansion */
            s_save(OP_DOMACRO, NIL, NIL);
            args = cons(code, NIL);
            code = value;
            s_goto(OP_APPLY);
        } else {
            code = cdr(code);
            s_goto(OP_E1ARGS);
        }

    case OP_E1ARGS:    /* eval arguments */
        args = cons(value, args);
        if (ispair(code)) {    /* continue */
            s_save(OP_E1ARGS, args, cdr(code));
            code = car(code);
            args = NIL;
            s_goto(OP_EVAL);
        } else {    /* end */
            args = reverse(args);
            code = car(args);
            args = cdr(args);
            s_goto(OP_APPLY);
        }

    case OP_APPLY:        /* apply 'code' to 'args' */
        if (isproc(code)) {
            s_goto(procnum(code));    /* PROCEDURE */
        } else if (isclosure(code)) {    /* CLOSURE */
            /* make environment */
            envir = cons(NIL, closure_env(code));
            for (x = car(closure_code(code)), y = args;
                 ispair(x); x = cdr(x), y = cdr(y)) {
                if (y == NIL) {
                    Error_0("Few arguments");
                } else {
                    car(envir) = cons(cons(car(x), car(y)), car(envir));
                }
            }
            if (x == NIL) {
                /*--
                 * if (y != NIL) {
                 *     Error_0("Many arguments");
                 * }
                 */
            } else if (issymbol(x))
                car(envir) = cons(cons(x, y), car(envir));
            else {
                Error_0("Syntax error in closure");
            }
            code = cdr(closure_code(code));
            args = NIL;
            s_goto(OP_BEGIN);
        } else if (iscontinuation(code)) {    /* CONTINUATION */
            dump = cont_dump(code);
            s_return(args != NIL ? car(args) : NIL);
        } else {
            Error_0("Illegal function");
        }

    case OP_DOMACRO:    /* do macro */
        code = value;
        s_goto(OP_EVAL);

    case OP_GENSYM:
        s_return(auto_gen());
    
    case OP_LAMBDA:    /* lambda */
        s_return(mk_closure(code, envir));

    case OP_QUOTE:        /* quote */
        s_return(car(code));

    case OP_DEF0:    /* define */
        if (ispair(car(code))) {
            x = caar(code);
            code = cons(LAMBDA, cons(cdar(code), cdr(code)));
        } else {
            x = car(code);
            code = cadr(code);
        }
        if (!issymbol(x)) {
            Error_0("Variable is not symbol");
        }
        s_save(OP_DEF1, NIL, x);
        s_goto(OP_EVAL);

    case OP_DEF1:    /* define */
        for (x = car(envir); x != NIL; x = cdr(x))
            if (caar(x) == code)
                break;
        if (x != NIL)
            cdar(x) = value;
        else
            car(envir) = cons(cons(code, value), car(envir));
        s_return(code);

    case OP_SET0:        /* set! */
        s_save(OP_SET1, NIL, car(code));
        code = cadr(code);
        s_goto(OP_EVAL);

    case OP_SET1:        /* set! */
        for (x = envir; x != NIL; x = cdr(x)) {
            for (y = car(x); y != NIL; y = cdr(y))
                if (caar(y) == code)
                    break;
            if (y != NIL)
                break;
        }
        if (x != NIL) {
            cdar(y) = value;
            s_return(value);
        } else {
            Error_1("Unbounded variable", code);
        }

    case OP_BEGIN:        /* begin */
        if (!ispair(code)) {
            s_return(code);
        }
        if (cdr(code) != NIL) {
            s_save(OP_BEGIN, NIL, cdr(code));
        }
        code = car(code);
        s_goto(OP_EVAL);

    case OP_IF0:        /* if */
        s_save(OP_IF1, NIL, cdr(code));
        code = car(code);
        s_goto(OP_EVAL);

    case OP_IF1:        /* if */
        if (istrue(value))
            code = car(code);
        else
            code = cadr(code);    /* (if #f 1) ==> () because
                         * car(NIL) = NIL */
        s_goto(OP_EVAL);

    case OP_LET0:        /* let */
        args = NIL;
        value = code;
        code = issymbol(car(code)) ? cadr(code) : car(code);
        s_goto(OP_LET1);

    case OP_LET1:        /* let (caluculate parameters) */
        args = cons(value, args);
        if (ispair(code)) {    /* continue */
            s_save(OP_LET1, args, cdr(code));
            code = cadar(code);
            args = NIL;
            s_goto(OP_EVAL);
        } else {    /* end */
            args = reverse(args);
            code = car(args);
            args = cdr(args);
            s_goto(OP_LET2);
        }

    case OP_LET2:        /* let */
        envir = cons(NIL, envir);
        for (x = issymbol(car(code)) ? cadr(code) : car(code), y = args;
             y != NIL; x = cdr(x), y = cdr(y))
            car(envir) = cons(cons(caar(x), car(y)), car(envir));
        if (issymbol(car(code))) {    /* named let */
            for (x = cadr(code), args = NIL; x != NIL; x = cdr(x))
                args = cons(caar(x), args);
            x = mk_closure(cons(reverse(args), cddr(code)), envir);
            car(envir) = cons(cons(car(code), x), car(envir));
            code = cddr(code);
            args = NIL;
        } else {
            code = cdr(code);
            args = NIL;
        }
        s_goto(OP_BEGIN);

    case OP_LET0AST:    /* let* */
        if (car(code) == NIL) {
            envir = cons(NIL, envir);
            code = cdr(code);
            s_goto(OP_BEGIN);
        }
        s_save(OP_LET1AST, cdr(code), car(code));
        code = cadaar(code);
        s_goto(OP_EVAL);

    case OP_LET1AST:    /* let* (make new frame) */
        envir = cons(NIL, envir);
        s_goto(OP_LET2AST);

    case OP_LET2AST:    /* let* (caluculate parameters) */
        car(envir) = cons(cons(caar(code), value), car(envir));
        code = cdr(code);
        if (ispair(code)) {    /* continue */
            s_save(OP_LET2AST, args, code);
            code = cadar(code);
            args = NIL;
            s_goto(OP_EVAL);
        } else {    /* end */
            code = args;
            args = NIL;
            s_goto(OP_BEGIN);
        }
    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);
    }
    return T;
}


pointer opexe_1(short op)
{
     pointer x, y;

    switch (op) {
    case OP_LET0REC:    /* letrec */
        envir = cons(NIL, envir);
        args = NIL;
        value = code;
        code = car(code);
        s_goto(OP_LET1REC);

    case OP_LET1REC:    /* letrec (caluculate parameters) */
        args = cons(value, args);
        if (ispair(code)) {    /* continue */
            s_save(OP_LET1REC, args, cdr(code));
            code = cadar(code);
            args = NIL;
            s_goto(OP_EVAL);
        } else {    /* end */
            args = reverse(args);
            code = car(args);
            args = cdr(args);
            s_goto(OP_LET2REC);
        }

    case OP_LET2REC:    /* letrec */
        for (x = car(code), y = args; y != NIL; x = cdr(x), y = cdr(y))
            car(envir) = cons(cons(caar(x), car(y)), car(envir));
        code = cdr(code);
        args = NIL;
        s_goto(OP_BEGIN);

    case OP_COND0:        /* cond */
        if (!ispair(code)) {
            Error_0("Syntax error in cond");
        }
        s_save(OP_COND1, NIL, code);
        code = caar(code);
        s_goto(OP_EVAL);

    case OP_COND1:        /* cond */
        if (istrue(value)) {
            if ((code = cdar(code)) == NIL) {
                s_return(value);
            }
            s_goto(OP_BEGIN);
        } else {
            if ((code = cdr(code)) == NIL) {
                s_return(NIL);
            } else {
                s_save(OP_COND1, NIL, code);
                code = caar(code);
                s_goto(OP_EVAL);
            }
        }

    case OP_DELAY:        /* delay */
        x = mk_closure(cons(NIL, code), envir);
        setpromise(x);
        s_return(x);

    case OP_AND0:        /* and */
        if (code == NIL) {
            s_return(T);
        }
        s_save(OP_AND1, NIL, cdr(code));
        code = car(code);
        s_goto(OP_EVAL);

    case OP_AND1:        /* and */
        if (isfalse(value)) {
            s_return(value);
        } else if (code == NIL) {
            s_return(value);
        } else {
            s_save(OP_AND1, NIL, cdr(code));
            code = car(code);
            s_goto(OP_EVAL);
        }

    case OP_OR0:        /* or */
        if (code == NIL) {
            s_return(F);
        }
        s_save(OP_OR1, NIL, cdr(code));
        code = car(code);
        s_goto(OP_EVAL);

    case OP_OR1:        /* or */
        if (istrue(value)) {
            s_return(value);
        } else if (code == NIL) {
            s_return(value);
        } else {
            s_save(OP_OR1, NIL, cdr(code));
            code = car(code);
            s_goto(OP_EVAL);
        }

    case OP_C0STREAM:    /* cons-stream */
        s_save(OP_C1STREAM, NIL, cdr(code));
        code = car(code);
        s_goto(OP_EVAL);

    case OP_C1STREAM:    /* cons-stream */
        args = value;    /* save value to register args for gc */
        x = mk_closure(cons(NIL, code), envir);
        setpromise(x);
        s_return(cons(args, x));

    case OP_0MACRO:    /* macro */
        x = car(code);
        code = cadr(code);
        if (!issymbol(x)) {
            Error_0("Variable is not symbol");
        }
        s_save(OP_1MACRO, NIL, x);
        s_goto(OP_EVAL);

    case OP_1MACRO:    /* macro */
        type_(value) |= T_MACRO;
        for (x = car(envir); x != NIL; x = cdr(x))
            if (caar(x) == code)
                break;
        if (x != NIL)
            cdar(x) = value;
        else
            car(envir) = cons(cons(code, value), car(envir));
        s_return(code);

    case OP_CASE0:        /* case */
        s_save(OP_CASE1, NIL, cdr(code));
        code = car(code);
        s_goto(OP_EVAL);

    case OP_CASE1:        /* case */
        for (x = code; x != NIL; x = cdr(x)) {
            if (!ispair(y = caar(x)))
                break;
            for ( ; y != NIL; y = cdr(y))
                if (eqv(car(y), value))
                    break;
            if (y != NIL)
                break;
        }
        if (x != NIL) {
            if (ispair(caar(x))) {
                code = cdar(x);
                s_goto(OP_BEGIN);
            } else {/* else */
                s_save(OP_CASE2, NIL, cdar(x));
                code = caar(x);
                s_goto(OP_EVAL);
            }
        } else {
            s_return(NIL);
        }

    case OP_CASE2:        /* case */
        if (istrue(value)) {
            s_goto(OP_BEGIN);
        } else {
            s_return(NIL);
        }
    case OP_PAPPLY:    /* apply */
        code = car(args);
        args = cadr(args);
        s_goto(OP_APPLY);

    case OP_PEVAL:    /* eval */
        code = car(args);
        args = NIL;
        s_goto(OP_EVAL);

    case OP_CONTINUATION:    /* call-with-current-continuation */
        code = car(args);
        args = cons(mk_continuation(dump), NIL);
        s_goto(OP_APPLY);

    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);
    }
    return T;
}


pointer opexe_2(short op)
{
     pointer x, y;
     long v;

    switch (op) {
    case OP_ADD:        /* + */
        for (x = args, v = 0; x != NIL; x = cdr(x))
            v += ivalue(car(x));
        s_return(mk_number(v));

    case OP_SUB:        /* - */
        for (x = cdr(args), v = ivalue(car(args)); x != NIL; x = cdr(x))
            v -= ivalue(car(x));
        s_return(mk_number(v));

    case OP_MUL:        /* * */
        for (x = args, v = 1; x != NIL; x = cdr(x))
            v *= ivalue(car(x));
        s_return(mk_number(v));

    case OP_DIV:        /* / */
        for (x = cdr(args), v = ivalue(car(args)); x != NIL; x = cdr(x)) {
            if (ivalue(car(x)) != 0)
                v /= ivalue(car(x));
            else {
                Error_0("Divided by zero");
            }
        }
        s_return(mk_number(v));

    case OP_REM:        /* remainder */
        for (x = cdr(args), v = ivalue(car(args)); x != NIL; x = cdr(x)) {
            if (ivalue(car(x)) != 0)
                v %= ivalue(car(x));
            else {
                Error_0("Divided by zero");
            }
        }
        s_return(mk_number(v));

    case OP_CAR:        /* car */
        if (ispair(car(args))) {
            s_return(caar(args));
        } else {
            Error_0("Unable to car for non-cons cell");
        }

    case OP_CDR:        /* cdr */
        if (ispair(car(args))) {
            s_return(cdar(args));
        } else {
            Error_0("Unable to cdr for non-cons cell");
        }

    case OP_CONS:        /* cons */
        cdr(args) = cadr(args);
        s_return(args);

    case OP_SETCAR:    /* set-car! */
        if (ispair(car(args))) {
            caar(args) = cadr(args);
            s_return(car(args));
        } else {
            Error_0("Unable to set-car! for non-cons cell");
        }

    case OP_SETCDR:    /* set-cdr! */
        if (ispair(car(args))) {
            cdar(args) = cadr(args);
            s_return(car(args));
        } else {
            Error_0("Unable to set-cdr! for non-cons cell");
        }

    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);
    }
    return T;
}


pointer opexe_3(short op)
{
     pointer x, y;

    switch (op) {
    case OP_NOT:        /* not */
        s_retbool(isfalse(car(args)));
    case OP_BOOL:        /* boolean? */
        s_retbool(car(args) == F || car(args) == T);
    case OP_NULL:        /* null? */
        s_retbool(car(args) == NIL);
    case OP_ZEROP:        /* zero? */
        s_retbool(ivalue(car(args)) == 0);
    case OP_POSP:        /* positive? */
        s_retbool(ivalue(car(args)) > 0);
    case OP_NEGP:        /* negative? */
        s_retbool(ivalue(car(args)) < 0);
    case OP_NEQ:        /* = */
        s_retbool(ivalue(car(args)) == ivalue(cadr(args)));
    case OP_LESS:        /* < */
        s_retbool(ivalue(car(args)) < ivalue(cadr(args)));
    case OP_GRE:        /* > */
        s_retbool(ivalue(car(args)) > ivalue(cadr(args)));
    case OP_LEQ:        /* <= */
        s_retbool(ivalue(car(args)) <= ivalue(cadr(args)));
    case OP_GEQ:        /* >= */
        s_retbool(ivalue(car(args)) >= ivalue(cadr(args)));
    case OP_SYMBOL:    /* symbol? */
        s_retbool(issymbol(car(args)));
    case OP_NUMBER:    /* number? */
        s_retbool(isnumber(car(args)));
    case OP_STRING:    /* string? */
        s_retbool(isstring(car(args)));
    case OP_PROC:        /* procedure? */
        /*--
             * continuation should be procedure by the example
             * (call-with-current-continuation procedure?) ==> #t
                 * in R^3 report sec. 6.9
             */
        s_retbool(isproc(car(args)) || isclosure(car(args))
              || iscontinuation(car(args)));
    case OP_PAIR:        /* pair? */
        s_retbool(ispair(car(args)));
    case OP_EQ:        /* eq? */
        s_retbool(car(args) == cadr(args));
    case OP_EQV:        /* eqv? */
        s_retbool(eqv(car(args), cadr(args)));
    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);
    }
    return T;
}


pointer opexe_4(short op)
{
     pointer x, y;

    switch (op) {
    case OP_FORCE:        /* force */
        code = car(args);
        if (ispromise(code)) {
            args = NIL;
            s_goto(OP_APPLY);
        } else {
            s_return(code);
        }

    case OP_WRITE:        /* write */
        print_flag = 1;
        args = car(args);
        s_goto(OP_P0LIST);

    case OP_DISPLAY:    /* display */
        print_flag = 0;
        args = car(args);
        s_goto(OP_P0LIST);

    case OP_NEWLINE:    /* newline */
        fprintf(outfp, "\n");
        s_return(T);

    case OP_ERR0:    /* error */
        if (!isstring(car(args))) {
            Error_0("error -- first argument must be string");
        }
        tmpfp = outfp;
        outfp = stderr;
                if (all_errors_fatal) {
                        FatalError(strvalue(car(args)));
                }
        fprintf(outfp, "Error: ");
        fprintf(outfp, "%s", strvalue(car(args)));
        args = cdr(args);
        s_goto(OP_ERR1);

    case OP_ERR1:    /* error */
        fprintf(outfp, " ");
        if (args != NIL) {
            s_save(OP_ERR1, cdr(args), NIL);
            args = car(args);
            print_flag = 1;
            s_goto(OP_P0LIST);
        } else {
            fprintf(outfp, "\n");
            flushinput();
            outfp = tmpfp;
            s_goto(OP_T0LVL);
        }

    case OP_REVERSE:    /* reverse */
        s_return(reverse(car(args)));

    case OP_APPEND:    /* append */
        s_return(append(car(args), cadr(args)));

    case OP_PUT:        /* put */
        if (!hasprop(car(args)) || !hasprop(cadr(args))) {
            Error_0("Illegal use of put");
        }
        for (x = symprop(car(args)), y = cadr(args); x != NIL; x = cdr(x))
            if (caar(x) == y)
                break;
        if (x != NIL)
            cdar(x) = caddr(args);
        else
            symprop(car(args)) = cons(cons(y, caddr(args)),
                          symprop(car(args)));
        s_return(T);

    case OP_GET:        /* get */
        if (!hasprop(car(args)) || !hasprop(cadr(args))) {
            Error_0("Illegal use of get");
        }
        for (x = symprop(car(args)), y = cadr(args); x != NIL; x = cdr(x))
            if (caar(x) == y)
                break;
        if (x != NIL) {
            s_return(cdar(x));
        } else {
            s_return(NIL);
        }

    case OP_QUIT:        /* quit */
        return (NIL);

    case OP_GC:        /* gc */
        gc(&NIL, &NIL);
        s_return(T);

    case OP_GCVERB:        /* gc-verbose */
    {    int    was = gc_verbose;
        
        gc_verbose = (car(args) != F);
        s_retbool(was);
    }

    case OP_NEWSEGMENT:    /* new-segment */
        if (!isnumber(car(args))) {
            Error_0("new-segment -- argument must be number");
        }
        fprintf(outfp, "allocate %d new segments\n",
            alloc_cellseg((int) ivalue(car(args))));
        s_return(T);
    }
}


pointer opexe_5(short op)
{
    pointer x, y;

    switch (op) {
    /* ========== reading part ========== */
    case OP_RDSEXPR:
        switch (tok) {
        case TOK_EOF:
            s_return(EOF_OBJ);
            
        case TOK_COMMENT:
            while (inchar() != '\n')
                ;
            tok = token();
            s_goto(OP_RDSEXPR);
        case TOK_LPAREN:
            tok = token();
            if (tok == TOK_RPAREN) {
                s_return(NIL);
            } else if (tok == TOK_DOT) {
                Error_0("syntax error -- illegal dot expression");
            } else {
                s_save(OP_RDLIST, NIL, NIL);
                s_goto(OP_RDSEXPR);
            }
        case TOK_QUOTE:
            s_save(OP_RDQUOTE, NIL, NIL);
            tok = token();
            s_goto(OP_RDSEXPR);
        case TOK_BQUOTE:
            s_save(OP_RDQQUOTE, NIL, NIL);
            tok = token();
            s_goto(OP_RDSEXPR);
        case TOK_COMMA:
            s_save(OP_RDUNQUOTE, NIL, NIL);
            tok = token();
            s_goto(OP_RDSEXPR);
        case TOK_ATMARK:
            s_save(OP_RDUQTSP, NIL, NIL);
            tok = token();
            s_goto(OP_RDSEXPR);
        case TOK_ATOM:
            s_return(mk_atom(readstr("();\t\n ")));
        case TOK_DQUOTE:
            s_return(mk_string(readstrexp()));
        case TOK_SHARP:
            if ((x = mk_const(readstr("();\t\n "))) == NIL) {
                Error_0("Undefined sharp expression");
            } else {
                s_return(x);
            }
        default:
            Error_0("syntax error -- illegal token");
        }
        break;

    case OP_RDLIST:
        args = cons(value, args);
        tok = token();
        if (tok == TOK_COMMENT) {
            while (inchar() != '\n')
                ;
            tok = token();
        }
        if (tok == TOK_RPAREN) {
            s_return(non_alloc_rev(NIL, args));
        } else if (tok == TOK_DOT) {
            s_save(OP_RDDOT, args, NIL);
            tok = token();
            s_goto(OP_RDSEXPR);
        } else {
            s_save(OP_RDLIST, args, NIL);;
            s_goto(OP_RDSEXPR);
        }

    case OP_RDDOT:
        if (token() != TOK_RPAREN) {
            Error_0("syntax error -- illegal dot expression");
        } else {
            s_return(non_alloc_rev(value, args));
        }

    case OP_RDQUOTE:
        s_return(cons(QUOTE_, cons(value, NIL)));

    case OP_RDQQUOTE:
        s_return(cons(QQUOTE, cons(value, NIL)));

    case OP_RDUNQUOTE:
        s_return(cons(UNQUOTE, cons(value, NIL)));

    case OP_RDUQTSP:
        s_return(cons(UNQUOTESP, cons(value, NIL)));

    /* ========== printing part ========== */
    case OP_P0LIST:
        if (!ispair(args)) {
            printatom(args, print_flag);
            s_return(T);
        } else if (car(args) == QUOTE_ && ok_abbrev(cdr(args))) {
            fprintf(outfp, "'");
            args = cadr(args);
            s_goto(OP_P0LIST);
        } else if (car(args) == QQUOTE && ok_abbrev(cdr(args))) {
            fprintf(outfp, "`");
            args = cadr(args);
            s_goto(OP_P0LIST);
        } else if (car(args) == UNQUOTE && ok_abbrev(cdr(args))) {
            fprintf(outfp, ",");
            args = cadr(args);
            s_goto(OP_P0LIST);
        } else if (car(args) == UNQUOTESP && ok_abbrev(cdr(args))) {
            fprintf(outfp, ",@");
            args = cadr(args);
            s_goto(OP_P0LIST);
        } else {
            fprintf(outfp, "(");
            s_save(OP_P1LIST, cdr(args), NIL);
            args = car(args);
            s_goto(OP_P0LIST);
        }

    case OP_P1LIST:
        if (ispair(args)) {
            s_save(OP_P1LIST, cdr(args), NIL);
            fprintf(outfp, " ");
            args = car(args);
            s_goto(OP_P0LIST);
        } else {
            if (args != NIL) {
                fprintf(outfp, " . ");
                printatom(args, print_flag);
            }
            fprintf(outfp, ")");
            s_return(T);
        }

    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);

    }
    return T;
}


pointer opexe_6(short op)
{
    pointer x, y;
    long v;
    static long    w;
    char    buffer[32];

    switch (op) {
    case OP_LIST_LENGTH:    /* list-length */    /* a.k */
        for (x = car(args), v = 0; ispair(x); x = cdr(x))
            ++v;
        s_return(mk_number(v));
        
    case OP_ASSQ:        /* assq */    /* a.k */
        x = car(args);
        for (y = cadr(args); ispair(y); y = cdr(y)) {
            if (!ispair(car(y))) {
                Error_0("Unable to handle non pair element");
            }
            if (x == caar(y))
                break;
        }
        if (ispair(y)) {
            s_return(car(y));
        } else {
            s_return(F);
        }
        
    case OP_PRINT_WIDTH:    /* print-width */    /* a.k */
        w = 0;
        args = car(args);
        print_flag = -1;
        s_goto(OP_P0_WIDTH);
        
    case OP_P0_WIDTH:
        if (!ispair(args)) {
            w += printatom(args, print_flag);
            s_return(mk_number(w));
        } else if (car(args) == QUOTE_
               && ok_abbrev(cdr(args))) {
            ++w;
            args = cadr(args);
            s_goto(OP_P0_WIDTH);
        } else if (car(args) == QQUOTE
               && ok_abbrev(cdr(args))) {
            ++w;
            args = cadr(args);
            s_goto(OP_P0_WIDTH);
        } else if (car(args) == UNQUOTE
               && ok_abbrev(cdr(args))) {
            ++w;
            args = cadr(args);
            s_goto(OP_P0_WIDTH);
        } else if (car(args) == UNQUOTESP
               && ok_abbrev(cdr(args))) {
            w += 2;
            args = cadr(args);
            s_goto(OP_P0_WIDTH);
        } else {
            ++w;
            s_save(OP_P1_WIDTH, cdr(args), NIL);
            args = car(args);
            s_goto(OP_P0_WIDTH);
        }
        
    case OP_P1_WIDTH:
        if (ispair(args)) {
            s_save(OP_P1_WIDTH, cdr(args), NIL);
            ++w;
            args = car(args);
            s_goto(OP_P0_WIDTH);
        } else {
            if (args != NIL)
                w += 3 + printatom(args, print_flag);
            ++w;
            s_return(mk_number(w));
        }
        
    case OP_GET_CLOSURE:    /* get-closure-code */    /* a.k */
        args = car(args);
        if (args == NIL) {
            s_return(F);
        } else if (isclosure(args)) {
            s_return(cons(LAMBDA, closure_code(value)));
        } else if (ismacro(args)) {
            s_return(cons(LAMBDA, closure_code(value)));
        } else {
            s_return(F);
        }
    case OP_CLOSUREP:        /* closure? */
        /*
         * Note, macro object is also a closure.
         * Therefore, (closure? <#MACRO>) ==> #t
         */
        if (car(args) == NIL) {
            s_return(F);
        }
        s_retbool(isclosure(car(args)));

    case OP_MACROP:        /* macro? */
        if (car(args) == NIL) {
            s_return(F);
        }
        s_retbool(ismacro(car(args)));

    default:
        sprintf(strbuff, "%d is illegal operator", operator);
        Error_0(strbuff);
    }
    return T;    /* NOTREACHED */
}


pointer    (*dispatch_table[])() = {
    opexe_0,    /* OP_LOAD = 0, */
    opexe_0,    /* OP_T0LVL, */
    opexe_0,    /* OP_T1LVL, */
    opexe_0,    /* OP_READ, */
    opexe_0,    /* OP_VALUEPRINT, */
    opexe_0,    /* OP_EVAL, */
    opexe_0,    /* OP_E0ARGS, */
    opexe_0,    /* OP_E1ARGS, */
    opexe_0,    /* OP_APPLY, */
    opexe_0,    /* OP_DOMACRO, */
    opexe_0,    // OP_GENSYM    
    
    opexe_0,    /* OP_LAMBDA, */
    opexe_0,    /* OP_QUOTE, */
    opexe_0,    /* OP_DEF0, */
    opexe_0,    /* OP_DEF1, */
    opexe_0,    /* OP_BEGIN, */
    opexe_0,    /* OP_IF0, */
    opexe_0,    /* OP_IF1, */
    opexe_0,    /* OP_SET0, */
    opexe_0,    /* OP_SET1, */
    opexe_0,    /* OP_LET0, */
    opexe_0,    /* OP_LET1, */
    opexe_0,    /* OP_LET2, */
    opexe_0,    /* OP_LET0AST, */
    opexe_0,    /* OP_LET1AST, */
    opexe_0,    /* OP_LET2AST, */
    
    opexe_1,    /* OP_LET0REC, */
    opexe_1,    /* OP_LET1REC, */
    opexe_1,    /* OP_LETREC2, */
    opexe_1,    /* OP_COND0, */
    opexe_1,    /* OP_COND1, */
    opexe_1,    /* OP_DELAY, */
    opexe_1,    /* OP_AND0, */
    opexe_1,    /* OP_AND1, */
    opexe_1,    /* OP_OR0, */
    opexe_1,    /* OP_OR1, */
    opexe_1,    /* OP_C0STREAM, */
    opexe_1,    /* OP_C1STREAM, */
    opexe_1,    /* OP_0MACRO, */
    opexe_1,    /* OP_1MACRO, */
    opexe_1,    /* OP_CASE0, */
    opexe_1,    /* OP_CASE1, */
    opexe_1,    /* OP_CASE2, */
    
    opexe_1,    /* OP_PEVAL, */
    opexe_1,    /* OP_PAPPLY, */
    opexe_1,    /* OP_CONTINUATION, */
    
    opexe_2,    /* OP_ADD, */
    opexe_2,    /* OP_SUB, */
    opexe_2,    /* OP_MUL, */
    opexe_2,    /* OP_DIV, */
    opexe_2,    /* OP_REM, */
    opexe_2,    /* OP_CAR, */
    opexe_2,    /* OP_CDR, */
    opexe_2,    /* OP_CONS, */
    opexe_2,    /* OP_SETCAR, */
    opexe_2,    /* OP_SETCDR, */
    
    opexe_3,    /* OP_NOT, */
    opexe_3,    /* OP_BOOL, */
    opexe_3,    /* OP_NULL, */
    opexe_3,    /* OP_ZEROP, */
    opexe_3,    /* OP_POSP, */
    opexe_3,    /* OP_NEGP, */
    opexe_3,    /* OP_NEQ, */
    opexe_3,    /* OP_LESS, */
    opexe_3,    /* OP_GRE, */
    opexe_3,    /* OP_LEQ, */
    opexe_3,    /* OP_GEQ, */
    opexe_3,    /* OP_SYMBOL, */
    opexe_3,    /* OP_NUMBER, */
    opexe_3,    /* OP_STRING, */
    opexe_3,    /* OP_PROC, */
    opexe_3,    /* OP_PAIR, */
    opexe_3,    /* OP_EQ, */
    opexe_3,    /* OP_EQV, */
    
    opexe_4,    /* OP_FORCE, */
    opexe_4,    /* OP_WRITE, */
    opexe_4,    /* OP_DISPLAY, */
    opexe_4,    /* OP_NEWLINE, */
    opexe_4,    /* OP_ERR0, */
    opexe_4,    /* OP_ERR1, */
    opexe_4,    /* OP_REVERSE, */
    opexe_4,    /* OP_APPEND, */
    opexe_4,    /* OP_PUT, */
    opexe_4,    /* OP_GET, */
    opexe_4,    /* OP_QUIT, */
    opexe_4,    /* OP_GC, */
    opexe_4,    /* OP_GCVERB, */
    opexe_4,    /* OP_NEWSEGMENT, */
    
    opexe_5,    /* OP_RDSEXPR, */
    opexe_5,    /* OP_RDLIST, */
    opexe_5,    /* OP_RDDOT, */
    opexe_5,    /* OP_RDQUOTE, */
    opexe_5,    /* OP_RDQQUOTE, */
    opexe_5,    /* OP_RDUNQUOTE, */
    opexe_5,    /* OP_RDUQTSP, */
    opexe_5,    /* OP_P0LIST, */
    opexe_5,    /* OP_P1LIST, */
    
    opexe_6,    /* OP_LIST_LENGTH, */
    opexe_6,    /* OP_ASSQ, */
    opexe_6,    /* OP_PRINT_WIDTH, */
    opexe_6,    /* OP_P0_WIDTH, */
    opexe_6,    /* OP_P1_WIDTH, */
    opexe_6,    /* OP_GET_CLOSURE, */
    opexe_6,    /* OP_CLOSUREP, */
    opexe_6,    /* OP_MACROP, */
    
};


/* kernel of this intepreter */
pointer Eval_Cycle(short op)
{
    operator = op;
    for (;;)
        if ((*dispatch_table[operator])(operator) == NIL)
            return NIL;
}

/* ========== Initialization of internal keywords ========== */

void mk_syntax(unsigned short op, char *name)
{
    pointer x;

    x = cons(mk_string(name), NIL);
    type_(x) = (T_SYNTAX | T_SYMBOL);
    syntaxnum(x) = op;
    oblist = cons(x, oblist);
}

void mk_proc(unsigned short op, char *name)
{
    pointer x, y;

    x = mk_symbol(name);
    y = get_cell(&NIL, &NIL);
    type_(y) = (T_PROC | T_ATOM);
    ivalue(y) = (long) op;
    car(global_env) = cons(cons(x, y), car(global_env));
}


void init_vars_global()
{
    pointer x;

    /* init input/output file */
    infp = stdin;
    outfp = stdout;
    /* init NIL */
    type_(NIL) = (T_ATOM | MARK_);
    car(NIL) = cdr(NIL) = NIL;
    /* init T */
    type_(T) = (T_ATOM | MARK_);
    car(T) = cdr(T) = T;
    /* init F */
    type_(F) = (T_ATOM | MARK_);
    car(F) = cdr(F) = F;
     /* init EOF_OBJ */
    type_(EOF_OBJ) = (T_ATOM | MARK_);
    car(EOF_OBJ) = cdr(EOF_OBJ) = EOF_OBJ;      //0.2,8

    /* init global_env */
    global_env = cons(NIL, NIL);
    /* init else */
    x = mk_symbol("else");
    car(global_env) = cons(cons(x, T), car(global_env));
}


void init_syntax()
{
    /* init syntax */
    mk_syntax(OP_LAMBDA, "lambda");
    mk_syntax(OP_QUOTE, "quote");
    mk_syntax(OP_DEF0, "define");
    mk_syntax(OP_IF0, "if");
    mk_syntax(OP_BEGIN, "begin");
    mk_syntax(OP_SET0, "set!");
    mk_syntax(OP_LET0, "let");
    mk_syntax(OP_LET0AST, "let*");
    mk_syntax(OP_LET0REC, "letrec");
    mk_syntax(OP_COND0, "cond");
    mk_syntax(OP_DELAY, "delay");
    mk_syntax(OP_AND0, "and");
    mk_syntax(OP_OR0, "or");
    mk_syntax(OP_C0STREAM, "cons-stream");
    mk_syntax(OP_0MACRO, "macro");
    mk_syntax(OP_CASE0, "case");
}


void init_procs()
{
    /* init procedure */
    mk_proc(OP_PEVAL, "eval");
    mk_proc(OP_PAPPLY, "apply");
    mk_proc(OP_CONTINUATION, "call-with-current-continuation");
    mk_proc(OP_FORCE, "force");
    mk_proc(OP_CAR, "car");
    mk_proc(OP_CDR, "cdr");
    mk_proc(OP_CONS, "cons");
    mk_proc(OP_SETCAR, "set-car!");
    mk_proc(OP_SETCDR, "set-cdr!");
    mk_proc(OP_ADD, "+");
    mk_proc(OP_SUB, "-");
    mk_proc(OP_MUL, "*");
    mk_proc(OP_DIV, "/");
    mk_proc(OP_REM, "remainder");
    mk_proc(OP_NOT, "not");
    mk_proc(OP_BOOL, "boolean?");
    mk_proc(OP_SYMBOL, "symbol?");
    mk_proc(OP_NUMBER, "number?");
    mk_proc(OP_STRING, "string?");
    mk_proc(OP_PROC, "procedure?");
    mk_proc(OP_PAIR, "pair?");
    mk_proc(OP_EQV, "eqv?");
    mk_proc(OP_EQ, "eq?");
    mk_proc(OP_NULL, "null?");
    mk_proc(OP_ZEROP, "zero?");
    mk_proc(OP_POSP, "positive?");
    mk_proc(OP_NEGP, "negative?");
    mk_proc(OP_NEQ, "=");
    mk_proc(OP_LESS, "<");
    mk_proc(OP_GRE, ">");
    mk_proc(OP_LEQ, "<=");
    mk_proc(OP_GEQ, ">=");
    mk_proc(OP_READ, "read");
    mk_proc(OP_WRITE, "write");
    mk_proc(OP_DISPLAY, "display");
    mk_proc(OP_NEWLINE, "newline");
    mk_proc(OP_LOAD, "load");
    mk_proc(OP_ERR0, "error");
    mk_proc(OP_REVERSE, "reverse");
    mk_proc(OP_APPEND, "append");
    mk_proc(OP_PUT, "put");
    mk_proc(OP_GET, "get");
    mk_proc(OP_GC, "gc");
    mk_proc(OP_GCVERB, "gc-verbose");
    mk_proc(OP_NEWSEGMENT, "new-segment");
    mk_proc(OP_LIST_LENGTH, "list-length");    /* a.k */
    mk_proc(OP_ASSQ, "assq");    /* a.k */
    mk_proc(OP_PRINT_WIDTH, "print-width");    /* a.k */    
    mk_proc(OP_GET_CLOSURE, "get-closure-code");    /* a.k */
    mk_proc(OP_CLOSUREP, "closure?");    /* a.k */
    mk_proc(OP_MACROP, "macro?");    /* a.k */
	mk_proc(OP_GENSYM, "gensym");
    mk_proc(OP_QUIT, "quit");
}


/* initialize several globals */
void init_globals()
{
    init_vars_global();
    init_syntax();
    init_procs();
    /* intialization of global pointers to special symbols */
    LAMBDA = mk_symbol("lambda");
    QUOTE_ = mk_symbol("quote");
    QQUOTE = mk_symbol("quasiquote");
    UNQUOTE = mk_symbol("unquote");
    UNQUOTESP = mk_symbol("unquote-splicing");
}

/* ========== Error ==========  */
void FatalError(char *s)
{
	fprintf(stderr, "Fatal error: %s\n", s);
	flushinput();
	args = NIL;

    exit(1);
}


/* ========== Main ========== */

void main()
{
    short op = (short) OP_LOAD;

    printf(banner);

    init_scheme();
    args = cons(mk_string(InitFile), NIL);
    Eval_Cycle(op);
    exit(0);
}

